﻿//*************************************************************************************************
//【サンプルプログラム】『Kinectから骨格情報を取得する』
//　DepthImage、ColorImageの人物切り取り表示は行わない
//【名称】Bitmap_GetSkeletonData　　【Version】1.02
//【作成者】私設研究所Neo-Tech-Lab.com  上田智章
//【作成日】2012年3月5日　【コメント追記】2012年6月30日
//【注意事項】
//　このプログラムをKinect for Windows SDK Version1.5で動かすには、参照設定の
// Microsoft.Kinectを一旦削除し、プロジェクト(P)-参照の追加(F)で、.NETタブ内
// にあるMicrosoft.Kinectを選択・追加してください。 
//*************************************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;                 //【Bitmap】【Color】
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;         //【BitmapData】
using System.Diagnostics;             //【Debug】
using System.Runtime.InteropServices; //【Marshal.Copy】
using Microsoft.Kinect;               //【参照設定】プロジェクト(P)⇒参照の追加(F)⇒.NETタブ⇒Microsoft.Kinectを選択

namespace Kinect_GestureControl
{
    public partial class Form1 : Form
    {
        //【KINECT関連】
        //　関数内でダイナミックに変数を確保すると動作速度が遅くなるので、ここにまとめて変数を確保しています。
        private KinectSensor KS_KinectSensor;   //【KINECT】
        private ColorImageFrame KS_ColorFrame;  //【KINECTのColorImageFrame】
        private byte[] KS_ColorImageDataARGB;   //【KINECTのColorImageData】32bit/画素,X8R8G8B8で受けてA8R8G8B8に変換
        private int KS_ColorFrameWidth;         //【KINECTのColorFrameの横方向画素数】
        private int KS_ColorFrameHeight;        //【KINECTのColorFrameの縦方向画素数】
        private Bitmap KS_bmp1;                 //【ColorImage表示用Bitmapオブジェクト】
        private BitmapData KS_data1;            //【ColorImage用BitmapData】
        private Rectangle KS_rect1;             //【ColorImage表示枠】
        private DepthImageFrame KS_DepthFrame;  //【KINECTのDepthImageFrame】
        private short[] KS_DepthData;           //【KINECTのDepthImageData】16bit/画素(MSB13bit:Depth, LSB3bit:PlayerID) 
        private byte[] KS_DepthImageDataXRGB;   //【KINECTのDepthImageData】32bit/画素,X8R8G8B8表示形式
        private int KS_DepthFrameWidth;         //【KINECTのDepthFrameの横方向画素数】
        private int KS_DepthFrameHeight;        //【KINECTのDepthFrameの縦方向画素数】
        private Bitmap KS_bmp2;                 //【DepthImage表示用Bitmapオブジェクト】
        private BitmapData KS_data2;            //【DepthImage用BitmapData】
        private Rectangle KS_rect2;             //【DepthImage表示枠】
        private int KS_PlayerIndexBitmask;      //【DepthデータのPlayer情報】 0000 0000 0000 0111B = 7
        private int KS_PlayerIndexBitmaskWidth; //【DepthデータのPlayer情報】 3 bits
        private int KS_Lmin;                    //【深度測定可能最近距離】default:  800mm, near: 400mm 
        private int KS_Lmax;                    //【深度測定可能最遠距離】default: 4000mm, near:3000mm
        private SkeletonFrame KS_SkeletonFrame; //【KINECTのSkeletonFrame】
        private Skeleton[] KS_SkeletonData;     //【KINECTのSkeleton】Skeleton 最大6名、20Joints/Player
        private bool KS_flagFlip = true;        //【左右反転フラグ】
        private bool KS_flagPlayer = false;     //【Player検出フラグ】
        private long KS_TimeStamp;              // ImageFrameのタイムスタンプ

        //【NTL拡張現実センサ関連】---------------------------------------------------------
        private NTL_ColorBar ColorBar = new NTL_ColorBar(); //【デプス値をカラーに変換する部品】
        private NTL_SkeletonData NTL_SkeletonData;
        private const int nFIFO = 60; //【ジェスチャー推定処理用FIFOの段数】2秒(60フレーム)分
        private SkeletonPoint Ps;   //working用
        private ColorImagePoint Pc; //working用
        private NTL_EstimationCircle estR = new NTL_EstimationCircle();
        private NTL_EstimationCircle estL = new NTL_EstimationCircle();
        private NTL_EstimationCircle.NTL_FigureType FigR, FigL; //検出状態
        private int DetectCountR = 0; //一瞬のLine判定の後、５秒間は無視。
        private int DetectCountL = 0; //一瞬のLine判定の後、５秒間は無視。
        private Random NTL_Rnd = new Random(); //【動画アクセス用乱数発生クラス】

        //【USB ADC基板のコントロール】
        private String COMPORT; //【A/Dコンバータ通信用ComPort番号】←ここはデバイス検出状況に応じて変更してください。
        System.IO.Ports.SerialPort SerialPort_for_ADC = new System.IO.Ports.SerialPort(); //【ADC用シリアルポート】
        private string RxMsg; //【受信文字列】


        public Form1()
        {
            InitializeComponent();

            //***********************************************************
            //【Kinect検出】【共通】
            //***********************************************************
            //【利用可能なKINECTセンサを取得する】
            KS_KinectSensor = (from sensorToCheck
                               in KinectSensor.KinectSensors
                               where sensorToCheck.Status == KinectStatus.Connected
                               select sensorToCheck).FirstOrDefault();
            //【検出結果の表示】
            if (KS_KinectSensor == null) //【Kinect接続が確認できなかった場合】
            {
                pictureBox2.Image = imageList1.Images[1];
                MessageBox.Show(
                   "接続可能なKINECTセンサが見当たりません。",
                   "KINECT for Windows Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            else //【Kinect検出に成功】
            {
                pictureBox2.Image = imageList1.Images[0];
            }

            //***********************************************************
            //【Kinectの初期化処理】
            //***********************************************************
            //【KINECTの設定】ColorStream, DepthStream, SkeletonStreamを許可
            KS_KinectSensor.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30); //【ColorStream許可】
            KS_KinectSensor.DepthStream.Enable(DepthImageFormat.Resolution320x240Fps30);    //【DepthStream許可】
            KS_KinectSensor.SkeletonStream.Enable();                                        //【SkeltonStream許可】
            KS_KinectSensor.Start();                                                        //【KINECTスタート】

            //【KINECTデータ取得イベントの設定とデータバッファの確保】
            KS_KinectSensor.ColorFrameReady += new EventHandler<ColorImageFrameReadyEventArgs>(ColorImageReady); //【カラー画像取得準備完了時のcall先】
            KS_KinectSensor.DepthFrameReady += new EventHandler<DepthImageFrameReadyEventArgs>(DepthImageReady); //【奥行き情報取得準備完了時のcall先】
            KS_KinectSensor.AllFramesReady += new EventHandler<AllFramesReadyEventArgs>(KinectAllFramesReady);   //【骨格情報取得準備完了時のcall先】
            KS_ColorImageDataARGB = new byte[KS_KinectSensor.ColorStream.FramePixelDataLength]; //【カラー画像データ用】
            KS_ColorFrameWidth = KS_KinectSensor.ColorStream.FrameWidth;   //【ColorImageの横方向画素数】
            KS_ColorFrameHeight = KS_KinectSensor.ColorStream.FrameHeight; //【ColorImageの縦方向画素数】
            KS_bmp1 = new Bitmap(KS_ColorFrameWidth, KS_ColorFrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【ColorImage表示用】
            KS_rect1 = new Rectangle(0, 0, KS_ColorFrameWidth, KS_ColorFrameHeight); //【ColorImage用枠】

            KS_DepthData = new short[KS_KinectSensor.DepthStream.FramePixelDataLength];              //【深度情報データ用】
            KS_DepthImageDataXRGB = new byte[KS_KinectSensor.DepthStream.FramePixelDataLength * 4];  //【深度情報データ表示用】
            KS_DepthFrameWidth = KS_KinectSensor.DepthStream.FrameWidth;   //【DepthImageの横方向画素数】
            KS_DepthFrameHeight = KS_KinectSensor.DepthStream.FrameHeight; //【DepthImageの縦方向画素数】
            KS_bmp2 = new Bitmap(KS_DepthFrameWidth, KS_DepthFrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【DepthImage表示用】
            KS_rect2 = new Rectangle(0, 0, KS_DepthFrameWidth, KS_DepthFrameHeight); //【DepthImage用枠】
            KS_PlayerIndexBitmask = DepthImageFrame.PlayerIndexBitmask;
            KS_PlayerIndexBitmaskWidth = DepthImageFrame.PlayerIndexBitmaskWidth;
            KS_KinectSensor.DepthStream.Range = DepthRange.Default;    //【Depth測定モードをDefaultに設定】
            KS_CheckDistance();                                        //【設定結果を表示】Lmax, Lminを取得

            KS_SkeletonData = new Skeleton[KS_KinectSensor.SkeletonStream.FrameSkeletonArrayLength]; //【骨格情報用】

            KS_KinectSensor.ElevationAngle = 0; //【Kinectヘッドの仰角を設定】設定範囲は±27°

            ColorBar.CreateColorPalletForAbsolute(KS_Lmax - KS_Lmin + 10); //【カラーパレットを作成する】
            pictureBox3.Image = ColorBar.CreateColorScale(256, 10);    //【256×10画素サイズのカラーバーを作成して表示する】

            pictureBox4.SizeMode = PictureBoxSizeMode.StretchImage; //【デプス・イメージがpictureBox枠内にはめ込まれます】
            pictureBox1.SizeMode = PictureBoxSizeMode.StretchImage; //【キャプチャー画像がpictureBox枠内にはめ込まれます】
            estR.Initialize(nFIFO); //【右手の軌跡から円を推定する準備を行う】
            estL.Initialize(nFIFO); //【左手の軌跡から円を推定する準備を行う】
            NTL_SkeletonData = new NTL_SkeletonData();

            string[] pn = System.IO.Ports.SerialPort.GetPortNames();
            if (pn.Length == 0)
            {
                MessageBox.Show(
                   "残念ながら、接続可能なシリアルポートが見つかりません。",
                   "USB ADC Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            COMPORT = pn[0]; //【一つしかシリアルポートが接続されていない前提で選択】

            InitializeSerialPort(COMPORT);
            ADCInitialize();

           
    //        YouTube(); //【YouTubeランダム】
    //        GoogleTTS("ja", "こんにちわ。しせつ研究所、ネオテックラボのキネクトワールドにようこそ。このプログラムでは、キネクトセンサを使って、ジェスチャー検出によるユーチューブ　どうがの　切り替え制御をおこないます。");
        }

        //***************************************************************
        //【Google Text-To-Speech】
        //***************************************************************
        //【日本語】ja 【英語】en 【ドイツ語】de 【フランス語】fr 【イタリア語】it 【スペイン語】es 【ポルトガル語】pt
        //【ポーランド語】pl 【ロシア語】ru //⇒×なぜか使えない【中国語(簡体字)】zh_CN 【中国語(繁体字)】zh_TW 【韓国語】ko
        private void GoogleTTS(string Language, string Message)
        {
            const string GoogleTTS_URL = "http://translate.google.com/translate_tts";
            webBrowser2.Url = new Uri(GoogleTTS_URL + "?tl=" + Language + "&q=" + Message);
        }

        //***************************************************************
        //【YouTube Access HTML】
        //***************************************************************
        private void YouTube()
        {
            //【YouTubeの自動Script再生方法については下記URLのJavaScriptを参考にしてください】
            const string Site1 = "http://www.geocities.jp/ivoiceapi/MovieViewer.htm";
            //【YouTube動画IDを登録します】
            string[] MovieIDList = { 
                //【ＩＡ】
                "PUr15HBFVWE", //★【ＩＡ】メテオ【オリジナルPV】
                "HE1sJfLkVjk", //★【MMD】 galaxias! ～IA～
                "R5oekmftgMc", //★【再】ＩＡさん達にgalaxias!を踊ってもらいました。[HD1080p]
                "6V_N3vAnauQ", //【VOCALOID MMD】Gravity=Reality【IA Model】
                "0JglNR2k36o", //【VOCALOID IA】 Little HERO!! 【オリジナル】
                //【TDA アペMiku】
                "KIdIzbnIAls", //【MMD】FREELY TOMORROW「初音ミク」
                "eHxmbhTaBxA", //★【MMD】Galaxias!「Newcomer: TDA Append Miku」
                "tVu_F8DmtRM", //★【MMD】【HD】 スイートマジック ★ Sweet Magic ☆ 初音ミク・アペンド
                "Wb65oGczTLc", //★[全面改修]【イロイロすげぇ】Tda式アペンドミク『FREELY TOMORROW』HD
                "cJLLzOSpEiU", //【MMD】恋はきっと急上昇☆【Tda式Appendミク】HD版（乳微揺れあり）
                //【らぶ式】
                "paCPNaPJjFw", //★【MMD】LOVE式ミク過激白泳裝Sweet Devil
                "lsXYhK-ARfs", //★【MMD】らぶ式ネルさんとミクさんが「Yellow」踊ってくれました。
                //【時報ちゃん】
                "9-b6LhVfS7w", //【MMD】時報ちゃんでSweetDevil
                "H1uhC3bZ6Jg", //【MMD】 トリノコシティ 【FSS】 (時報ちゃん＆ISAOモデルミクact2改)[HD1080p]
                "hh0HQvat6Y4", //【MMD-PV】時報ちゃん Project Dance Ver. [HD1080p](日本語歌詞付)
                "2pNp6Am62wc", //【MMD】時報ちゃん 『プラチナ』-shin'in future Mix-【PV】[HD1080p](日本語歌詞)
                //【夏海ほかさぼてんモデル】
                "_yZ1shzn_so", //★【MMD】さぼてんGUMIと夏海のYOUTHFUL DAYS' GRAFFITI（MME）
                "jPsdlwv-q90", //【MMD】夏海お姉ちゃんと一緒に 「エレクトリック・ラブ」
                "CCW18WvVb3I", //MMD DMC2 マトリョシカ
                //【Piron式】
                "ndSrA9Fi4Lw", //【MMD】Pironモデル ミク・メイコ・ルカで「BREEZE」
                "PNFzPBpL46o", //【MMD】Piron式ミクさんにトリノコシティを躍ってもらった
                "VN6_W2S-x3Y", //【MMD】きゅんきゅんメイズに『Prism Heart』を踊ってもらいました
                "SKCX_Zchd8U", //【MMD】きゅんきゅんメイドたちで『ハッピーシンセサイザ』
                //【LAT式】
                "hAkQ1zZJI1c", //★【MMD】Lat式ミク(v2.2)で「教えて!!魔法のLyric」(HD)
                "AQlbraKK5yo", //★[MMD Cup 7] あざといビーム【脱衣もあるよ！】 [Hatsune Miku]
                "aKv7sp4NhsU", //[MMD] 初音ミク 千本桜
                "qJ5UBvr8F14", //★【MMD】千本桜【初音ミク】
                "zrigiQs7YPY", //【MMD】おちゃめ機能Full【HD720P】
                "WH8L226nOHU", //【初音ミク】Downloader【オリジナル曲+PV】
                "Gp0zyjz_YnQ", //【MMD】Pironモデルに『EX-GIRL』を踊ってもらいました。
                "MfnFKZ5nBQU", //MMD Packaged Lat Model Miku
                //【かにひらモデル】
                "lOY4CB0_AOI", //[MMD] Gravity=Reality (REUPLOAD)
                "YSYtL4YyGa0", //【MMD】レアさんにミニスカで踊ってもらう夢を見た
                "oM8y6zMLWMY", //【MMD】レア様が水着に着替えたら...【ねこみみスイッチ】 [HD1080p]
                "57VFO1-tUWU", //【MMD】レアさんでエレクトリック・ラブ
                //【SEGA】
                "ZUeh21LsWg0", //★【初音ミク】初音ミク「ローリンガール」を少しだけ【Project DIVA extend】
                "wmcxyQNKt3E", //「初音ミクの消失」PV ディレクターズカット版
                //【初音ミク他ボカロ】
                "QhdPXL1cJq8", //【第7回MMD杯本選】Bad Apple!! feat.Miku Hatsune &amp; Kasane Teto
                "n-yFoLcwtrU", //MMD ハピーシンセイザあにまさVER
                "f4bNVB5D5m4", //初音ミク YOUTHFUL DAYS' GRAFFTI
                "kcYzyDKipz0", //-VOCALOID- 【初音ミク Hatune Miku】 The secret garden
                "ui6YatIwU2A", //【MikuMikuDance】「Packaged」Act.3【PV】【HD 720p】+mp3♫
                "CUrNY2wseRo", //【初音ミク】可能世界のロンド【3DPV】
                "iCOfYnPdOcI", //Miku Hatsune - Uninstall
                "1Ky7Y4-6qAM", //【初音ミク】LOL -lots of laugh-【オリジナル】
                "PqJNc9KVIZE", //livetune feat. 初音ミク 『Tell Your World』Music Video
                "e9Bs7UbwU28", //【猫村いろはオリジナル】 恋のI・RO・HA教えます 【MMD-PV】
                "vMp2zn-wBxQ", //【蒼姫ラピス】Think the Future【オリジナル曲】
                "0WfVWB5rr3I", //MMD MIKI MIKI DANCE
                //【夏海ほかさぼてんモデル】
                "_yZ1shzn_so", //★【MMD】さぼてんGUMIと夏海のYOUTHFUL DAYS' GRAFFITI（MME）
                //【らぶ式】
                "paCPNaPJjFw", //★【MMD】LOVE式ミク過激白泳裝Sweet Devil
                //【ＩＡ】
                "PUr15HBFVWE", //★【ＩＡ】メテオ【オリジナルPV】
                "HE1sJfLkVjk", //★【MMD】 galaxias! ～IA～
                "R5oekmftgMc", //★【再】ＩＡさん達にgalaxias!を踊ってもらいました。[HD1080p]
                "6V_N3vAnauQ", //【VOCALOID MMD】Gravity=Reality【IA Model】
                //【TDA アペMiku】
                "KIdIzbnIAls", //【MMD】FREELY TOMORROW「初音ミク」
                //【LAT式】
                "hAkQ1zZJI1c", //★【MMD】Lat式ミク(v2.2)で「教えて!!魔法のLyric」(HD)
                "AQlbraKK5yo", //★[MMD Cup 7] あざといビーム【脱衣もあるよ！】 [Hatsune Miku]
                "CleQxLRpV-I"  //[MMD] Miki Miki ★ Romantic Night [MMD CUP IV]
                                   };

            int RN = (int)(NTL_Rnd.NextDouble()*(float)MovieIDList.Length);
            webBrowser1.Url = new Uri(Site1 + "?" + MovieIDList[RN]); //【YouTube動画ID】上のMovieIDListからランダムに選択
        }

        //***************************************************************
        //【CheckDistance】Defaultモード(800mm～4000mm)
        //***************************************************************
        private void radioButton1_CheckedChanged(object sender, EventArgs e)
        {  //【Defaultモード】
            KS_KinectSensor.DepthStream.Range = DepthRange.Default;
            KS_CheckDistance();
        }

        //***************************************************************
        //【距離モード変更】Nearモード(400m～3000mm)
        //***************************************************************
        private void radioButton2_CheckedChanged(object sender, EventArgs e)
        {  //【Nearモード】
            KS_KinectSensor.DepthStream.Range = DepthRange.Near;
            KS_CheckDistance();
        }

        //***************************************************************
        //【CheckDistance】KINECTセンサの検出可能距離現在値を調べる
        //***************************************************************
        private void KS_CheckDistance()
        {
            KS_Lmax = KS_KinectSensor.DepthStream.MaxDepth;
            KS_Lmin = KS_KinectSensor.DepthStream.MinDepth;
            label1.Text = "Min=" + KS_Lmin.ToString() + "mm";
            label2.Text = "Max=" + KS_Lmax.ToString() + "mm";
        }

        //***************************************************************
        //【終了要求ボタン】
        //***************************************************************
        private void button1_Click(object sender, EventArgs e)
        {
            if (KS_KinectSensor != null)
            {
                KS_KinectSensor.Stop();    //【KINECTセンサを終了】★これを実行しないとレーザーが消えない
                KS_KinectSensor.Dispose(); //【KINECTセンサのオブジェクトを解放する】


                ADCTerminate();
                TerminateSerialPort();
            }
            this.Close();              //【Formを閉じる】
        }

        //***************************************************************
        //【上下角変更ボタン】KINECTの上下角を変更。±27°の範囲内。時間を置かないとエラーが発生する
        //***************************************************************
        private void button2_Click(object sender, EventArgs e)
        {
            Int32 angle = int.Parse(textBox1.Text); //【設定角度を読み込む】±27°
            if (angle <= KS_KinectSensor.MaxElevationAngle && angle >= KS_KinectSensor.MinElevationAngle)
            {
                KS_KinectSensor.ElevationAngle = angle; //【上下角の変更】
            }
        }

        //***************************************************************
        //【カラー画像の左右反転指定フラグ】
        //***************************************************************
        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked) { KS_flagFlip = true; } else { KS_flagFlip = false; }
        }

        //***************************************************************
        //【ColorImageの取得イベント】
        //***************************************************************
        private void ColorImageReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            using (KS_ColorFrame = e.OpenColorImageFrame())
            {
                if (KS_ColorFrame != null)
                {
                    int i, j, k1, k2, n, xc, yc, r;
                    Graphics g;
                    SolidBrush newbrush;
                    Pen cp;

                    KS_ColorFrame.CopyPixelDataTo(KS_ColorImageDataARGB); //【KINECTのColor画像を配列PixelDataにコピー】
                    for(j=3; j<KS_ColorImageDataARGB.Length; j+=4)
                    {
                        KS_ColorImageDataARGB[j] = 255;
                    }
                    KS_data1 = KS_bmp1.LockBits(KS_rect1, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lock】
                    Marshal.Copy(KS_ColorImageDataARGB, 0, KS_data1.Scan0, KS_ColorImageDataARGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
                    KS_bmp1.UnlockBits(KS_data1);           //【Lock解除】
                    if (KS_flagFlip) { KS_bmp1.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox1.Image = KS_bmp1;            //【pictureBox1に画像表示】

                    if (KS_flagPlayer == false) return;

                    g = Graphics.FromImage(pictureBox1.Image);

                    if (checkBox2.Checked) //【骨格表示処理】
                    {
                        newbrush = new SolidBrush(Color.Red);
                        for (j = 0; j < NTL_SkeletonData.JointPosition.Length; j++)
                        {
                            g.FillEllipse(newbrush, (float)NTL_SkeletonData.JointPosition[j].X - 5f, (float)NTL_SkeletonData.JointPosition[j].Y - 5f, 10f, 10f);
                        }
                        for (j = 0; j < NTL_SkeletonData.Bone.Length; j++)
                        {
                            k1 = NTL_SkeletonData.BoneIndexList[2 * j];
                            k2 = NTL_SkeletonData.BoneIndexList[2 * j + 1];
                            g.DrawLine(new Pen(Color.Red, 3.0f),
                                       new PointF((float)NTL_SkeletonData.JointPosition[k1].X, (float)NTL_SkeletonData.JointPosition[k1].Y),
                                       new PointF((float)NTL_SkeletonData.JointPosition[k2].X, (float)NTL_SkeletonData.JointPosition[k2].Y));
                        }
                    }

                    if (checkBox3.Checked) //【ジェスチャー推定処理】
                    {
                        if (estL.Full)
                        {
                            //【左手軌跡の推定円を描画】
                            xc = (int)estL.X; yc = (int)estL.Y; r = (int)estL.Radius;
                            cp = new Pen(Color.DeepSkyBlue, 5.0f);
                            g.DrawEllipse(cp, xc - r, yc - r, 2 * r, 2 * r);
                        }
                        if (estR.Full)
                        {
                            //【右手軌跡の推定円を描画】
                            xc = (int)estR.X; yc = (int)estR.Y; r = (int)estR.Radius;
                            cp = new Pen(Color.DeepPink, 5.0f);
                            g.DrawEllipse(cp, xc - r, yc - r, 2 * r, 2 * r);
                        }
                    }

                    //【右手の残像を描画】
                    newbrush = new SolidBrush(Color.Red);
                    if (estR.Full) { j = estR.Pointer; n = estR.Length; } else { j = 0; n = estR.Pointer; }
                    for (i = 0; i < n; i++)
                    {
                        g.FillEllipse(newbrush, estR.Px[j] - 5f, estR.Py[j] - 5f, 10f, 10f);
                        j++; if (j == n) { j = 0; }
                    }
                    //【左手の残像を描画】
                    newbrush = new SolidBrush(Color.Green);
                    if (estL.Full) { j = estL.Pointer; n = estL.Length; } else { j = 0; n = estL.Pointer; }
                    for (i = 0; i < n; i++)
                    {
                        g.FillEllipse(newbrush, estL.Px[j] - 5f, estL.Py[j] - 5f, 10f, 10f);
                        j++; if (j == n) { j = 0; }
                    }

                    if (checkBox3.Checked) //【ジェスチャー推定処理】
                    {
                        g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAlias;
                        string s;
                        s = "左手：" + CheckEstimationResult(estL.Type);
                        g.DrawString(s, new System.Drawing.Font("New Times Roman", 20.0F), new SolidBrush(Color.Green), 0, 0);
                        s = "右手：" + CheckEstimationResult(estR.Type);
                        g.DrawString(s, new System.Drawing.Font("New Times Roman", 20.0F), new SolidBrush(Color.Red), 0, 25);
                        if (estR.Type == NTL_EstimationCircle.NTL_FigureType.Circle)
                        {
                            if (DetectCountR == 0)
                            {
                             //   GoogleTTS("ja", "どうがをきりかえます");
                             //   YouTube(); //右手の『円』を検出した場合、YouTube動画切替
                                label4.Text = "X = " + estR.X.ToString();
                                label5.Text = "Y = " + estR.Y.ToString();
                                label6.Text = "r = " + estR.Radius.ToString();
                            } 
                            if (DetectCountR < 150) { DetectCountR++; } //【５秒間は不感】
                        }
                        else
                        {
                            DetectCountR = 0;
                        }

                        if ((estL.Type == NTL_EstimationCircle.NTL_FigureType.Circle) && (estL.Radius>50))
                        {
                            if (DetectCountL == 0)
                            {
                        //        GoogleTTS("ja", "エアコンを操作します");
                                label4.Text = "エアコンを操作します";
                                label5.Text = "";
                                label6.Text = "";
                                SetAirCon(); //【エアコン】
                            }
                            if (DetectCountL < 150) { DetectCountL++; } //【５秒間は不感】
                        }
                        if (estL.Type==NTL_EstimationCircle.NTL_FigureType.Point) 
                        {
                            DetectCountL = 0;
                        }


                    }

                    g.Dispose();
                    this.Validate();
                }
                else
                {
                    //【ImageFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }

        private string CheckEstimationResult(NTL_EstimationCircle.NTL_FigureType t)
        {
            string s = "";
            switch (t)
            {
                case NTL_EstimationCircle.NTL_FigureType.Circle:
                    s = "円が検出されました"; break;
                case NTL_EstimationCircle.NTL_FigureType.Arc:
                    s = "円弧が検出されました"; break;
                case NTL_EstimationCircle.NTL_FigureType.Line:
                    s = "線分が検出されました"; break;
                case NTL_EstimationCircle.NTL_FigureType.Point:
                    s = "点が検出されました"; break;
            }
            return s;
        }

        //***************************************************************
        //【DepthImageの取得イベント】
        //***************************************************************
        private void DepthImageReady(object sender, DepthImageFrameReadyEventArgs e)
        {
            using (KS_DepthFrame = e.OpenDepthImageFrame())
            {
                if (KS_DepthFrame != null)
                {
                    KS_DepthFrame.CopyPixelDataTo(KS_DepthData); //【KINECTのDepth情報を配列DepthDataにコピー】
                    KS_ConvertDepthToColor(); //【深度情報をカラーバーの色を使って表示形式をX8R8G8B8形式に変換】
                    KS_data2 = KS_bmp2.LockBits(KS_rect2, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lockする】
                    Marshal.Copy(KS_DepthImageDataXRGB, 0, KS_data2.Scan0, KS_DepthImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列depthDataをコピーする】
                    KS_bmp2.UnlockBits(KS_data2); //【Lock解除】
                    if (KS_flagFlip) { KS_bmp2.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox4.Image = KS_bmp2;  //【pictureBox1に画像表示】
                }
                else
                {
                    //【DepthFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }

        //***************************************************************
        //【ConvertDepthToColor】16ビットのDepth値をカラーバーに対応する色に変更
        //***************************************************************
        private void KS_ConvertDepthToColor()
        {
            float a;      //【Normalモード】800～4000mm ⇒ 0.0796875  3200色
            int i, j, d;  //【Nearモード】  400～3000mm ⇒ 0.0980769　2600色
            int f; //【PlayerLost検出用】
            int playerID; //【下位3bitはplayerID】
            if (KS_KinectSensor.DepthStream.Range == DepthRange.Default) { a = 1.000000f; } else { a = 1.230769f; }
            j = 0;
            f = 0;
            for (i = 0; i < KS_DepthData.Length; i++)
            {
                d = KS_DepthData[i];                  //【距離情報】【上位13bitがDepth値】
                playerID = d & KS_PlayerIndexBitmask; //【下位3bitはplayerID】【未使用】
                if (playerID > 0) { f = 1; }
                d = ((d >> KS_PlayerIndexBitmaskWidth) & 0x1FFF);  //【realDepth】0mm～8191mm
                if (d < KS_Lmin) { d = KS_Lmin; } //【近すぎる】
                if (d > KS_Lmax) { d = KS_Lmax; } //【遠すぎる】
                d = (int)((float)(d - KS_Lmin) * a) * 4;
                KS_DepthImageDataXRGB[j] = ColorBar.NTL_ColorElement[d]; //【青】B
                KS_DepthImageDataXRGB[j + 1] = ColorBar.NTL_ColorElement[d + 1]; //【緑】G
                KS_DepthImageDataXRGB[j + 2] = ColorBar.NTL_ColorElement[d + 2]; //【赤】R
                KS_DepthImageDataXRGB[j + 3] = ColorBar.NTL_ColorElement[d + 3]; //【α】
                j = j + 4;
            }
            if (f == 0) { KS_flagPlayer = false; } else { KS_flagPlayer = true; }
        }

        //***************************************************************
        //【SkeletonFrameの取得イベント】SkeletonFrameは6名分。各々がJointsコレクションを持つ。
        //***************************************************************
        private void KinectAllFramesReady(object sender, AllFramesReadyEventArgs e)
        {
            using (KS_SkeletonFrame = e.OpenSkeletonFrame())
            {
                if (KS_SkeletonFrame != null)
                {
                    KS_SkeletonFrame.CopySkeletonDataTo(KS_SkeletonData); //【KINECTのSkeltonデータを配列SkeletonDataにコピー】
                    KS_TimeStamp = KS_SkeletonFrame.Timestamp; //【現在時刻をミリ秒単位で取得】
                    for (int i = 0; i < KS_SkeletonData.Length; i++)
                    {
                        if (KS_SkeletonData[i].TrackingState == SkeletonTrackingState.Tracked)
                        {
                            label3.Text = "Tracking Player No.=" + i.ToString();
                            NTL_SkeletonData.SetBody(KS_KinectSensor, KS_SkeletonData[i], KS_flagFlip); //【骨格情報の保存】
                            if (KS_SkeletonData[i].Joints[JointType.HandRight].TrackingState == JointTrackingState.Tracked)
                            {
                                Ps = KS_SkeletonData[i].Joints[JointType.HandRight].Position; //【右手SkeletonPointを取得】
                                if (KS_flagFlip) { Ps.X = -(Ps.X); } //【必要なら左右を反転】
                                Pc = KS_KinectSensor.MapSkeletonPointToColor(Ps, ColorImageFormat.RgbResolution640x480Fps30);
                                estR.Estimation((float)Pc.X, (float)Pc.Y); //【右手軌跡を追加して円を推定】
                            }
                            if (KS_SkeletonData[i].Joints[JointType.HandLeft].TrackingState == JointTrackingState.Tracked)
                            {
                                Ps = KS_SkeletonData[i].Joints[JointType.HandLeft].Position;  //【左手SkeletonPointを取得】
                                if (KS_flagFlip) { Ps.X = -(Ps.X); } //【必要なら左右を反転】
                                Pc = KS_KinectSensor.MapSkeletonPointToColor(Ps, ColorImageFormat.RgbResolution640x480Fps30);
                                estL.Estimation((float)Pc.X, (float)Pc.Y); //【左手軌跡を追加して円を推定】
                            }
                            break;
                        }
                    }
                }
                else
                {
                    //【SkeletonFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }


        //***********************************************************************
        // 時間待ち(ミリ秒単位)
        //***********************************************************************
        [DllImport("kernel32.dll")]
        public static extern void Sleep(Int32 dwMilliseconds);

        public Boolean InitializeSerialPort(string com)
        {
            //【ADC用シリアルポートの設定】
            this.SerialPort_for_ADC.PortName = COMPORT;
            // トラ技増刊号の付録基板は疑似シリアルポート
            //     this.SerialPort_for_ADC.BaudRate = 115200;
            //     this.SerialPort_for_ADC.DataBits = 8;
            //     this.SerialPort_for_ADC.StopBits = System.IO.Ports.StopBits.One;
            //     this.SerialPort_for_ADC.Parity = System.IO.Ports.Parity.None;
            //     this.SerialPort_for_ADC.ReadTimeout = 50;
            //     this.SerialPort_for_ADC.ReadBufferSize = 256;
            //     this.SerialPort_for_ADC.WriteBufferSize = 256;
            try { this.SerialPort_for_ADC.Open(); }
            catch (Exception) { MessageBox.Show("シリアルポートが開けません"); return false; }
            return true;
        }

        public void TerminateSerialPort()
        {
            this.SerialPort_for_ADC.Close();
        }

        public void SetAirCon()
        {
            this.SerialPort_for_ADC.Write("PW01\r\n");  //【送信】bit0
            Sleep(50);
            RxMsg = this.SerialPort_for_ADC.ReadLine();
            Sleep(500);
            this.SerialPort_for_ADC.Write("PW00\r\n");  //【送信】bit0
            Sleep(50);
            RxMsg = this.SerialPort_for_ADC.ReadLine();
        }


        // ■■■■■■■■■■　【A/Dコンバータの初期化処理】　■■■■■■■■■■
        public void ADCInitialize()
        {
            //【接続確認】特に値のチェックなどはしていません。
            this.SerialPort_for_ADC.Write("E\r\n");       //【送信】"E"+CR+LF
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            //【ADCモードレジスタ書込み】000A　ADC変換レートを(16.7Hzに)設定する
            // this.SerialPort_for_ADC.Write("C080001\r\n"); //【送信】"C08"+"0001"+CR+LF 500Hz
            // this.SerialPort_for_ADC.Write("C080002\r\n"); //【送信】"C08"+"0002"+CR+LF 250Hz
            // this.SerialPort_for_ADC.Write("C080003\r\n"); //【送信】"C08"+"0003"+CR+LF 125Hz
            // this.SerialPort_for_ADC.Write("C080004\r\n"); //【送信】"C08"+"0004"+CR+LF 62.5Hz
            this.SerialPort_for_ADC.Write("C080005\r\n"); //【送信】"C08"+"0005"+CR+LF 50Hz
            // this.SerialPort_for_ADC.Write("C080006\r\n"); //【送信】"C08"+"0006"+CR+LF 39.2Hz
            // this.SerialPort_for_ADC.Write("C080007\r\n"); //【送信】"C08"+"0007"+CR+LF 33.3Hz
            // this.SerialPort_for_ADC.Write("C080008\r\n"); //【送信】"C08"+"0008"+CR+LF 19.6Hz
            // this.SerialPort_for_ADC.Write("C08000A\r\n"); //【送信】"C08"+"000A"+CR+LF 16.7Hz
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            //【ADC設定レジスタ書込み】5080 【コマンド】C105080 + CR + LF
            this.SerialPort_for_ADC.Write("C105080\r\n"); //【入力】チャンネル0
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            //【IOレジスタ】0A　【コマンド】C280A + CR + LF
            this.SerialPort_for_ADC.Write("C280A\r\n");   //【送信】電流源は未使用
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            //【ADCオフセットレジスタ】800000 【コマンド】C30800000 + CR + LF
            this.SerialPort_for_ADC.Write("C30800000\r\n"); //【送信】
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            this.SerialPort_for_ADC.Write("PM80\r\n"); //【送信】
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            this.SerialPort_for_ADC.Write("PW78\r\n"); //【送信】
            Sleep(3000);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            this.SerialPort_for_ADC.Write("PW00\r\n"); //【送信】
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            this.SerialPort_for_ADC.DiscardInBuffer();    //【バッファ内データを破棄する】
        }

        // ■■■■■■■■■■　【A/Dコンバータの終了処理】　■■■■■■■■■■
        public void ADCTerminate()
        {
            this.SerialPort_for_ADC.Write("C08600A\r\n"); //【送信】パワーダウンモード
            Sleep(5);                                     //【待機】5m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
        }


    }
}
